/*
 * cdebconf frontend, corba idl
 */

module Debconf {
        exception Failure {
                string reason;
        };
        
        interface Frontend {
                typedef sequence <string,65536> selects;
                
                boolean Boolean(in string prompt) raises (Failure);
                unsigned long long Multiselect(in unsigned short count,
                                               in selects choices)
                        raises (Failure);
                oneway void Note(in string note) raises (Failure);
                string Password(in string prompt) raises (Failure);
                unsigned short Select(in unsigned short count,
                                      in selects choices)
                        raises (Failure);
                string String(in string prompt) raises (Failure);
                string Text(in string prompt) raises (Failure);
        };
};

/*
 * Local Variables:
 * mode: c
 * tab-width: 8
 * indent-tabs-mode: nil
 * c-basic-offset: 8
 * End:
 */
